//***************************************************************************
//
//  Rupl.c
//
//***************************************************************************

#include <windows.h>
#include <shlobj.h>
#include <tlhelp32.h>
#include <winver.h>
#include <string.h>

__declspec (dllimport) BOOL SetHookHandle (HHOOK);
__declspec (dllimport) BOOL SetWindowHandle (HWND);
__declspec (dllimport) LRESULT CALLBACK ShellProc (int, WPARAM, LPARAM);

LRESULT CALLBACK WndProc (HWND, UINT, WPARAM, LPARAM);

void UpdateMRUList (HWND, HWND, LPSTR);
BOOL HwndToPath (HWND, LPSTR, int);
BOOL HwndToTitle (HWND, LPSTR, LPSTR, int);
BOOL ExeToTitle (LPSTR, LPSTR, int);
BOOL GetTitleFromRegistry (LPSTR, LPSTR, int);
BOOL RefreshExistingLnkFile (HWND, LPSTR);
BOOL IsExeOK (LPSTR);
int FindString (LPSTR, LPSTR);
BOOL ValidateTitle (LPSTR);
BOOL IsDuplicate (HWND, LPSTR);
BOOL GetProgramsFolder (HWND, LPSTR);
BOOL GetRecentlyUsedFolder (HWND, LPSTR);
HRESULT CreateLink (LPSTR, LPSTR, LPSTR);
HRESULT ResolveLink (LPSTR, LPSTR, int);
UINT GetLnkFileCount (LPSTR);
void DeleteOldestLnkFile (LPSTR);
void DeleteNonDescriptiveName (HWND, LPSTR);
void UpdateFileDateAndTime (LPSTR);
UINT GetMRUMax ();

/////////////////////////////////////////////////////////////////////////////
// Global data

char g_szRegKey[] = "Software\\PC Magazine\\RUPL\\V1.0";

/////////////////////////////////////////////////////////////////////////////
// WinMain

int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
    LPSTR lpszCmdLine, int nCmdShow)
{
    static char szAppName[] = "RUPL";
    WNDCLASS wc;
    HWND hwnd;
    MSG msg;

    wc.style = 0;
    wc.lpfnWndProc = (WNDPROC) WndProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hInstance = hInstance;
    wc.hIcon = LoadIcon (NULL, IDI_APPLICATION);
    wc.hCursor = LoadCursor (NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH) (COLOR_WINDOW + 1);
    wc.lpszMenuName = NULL;
    wc.lpszClassName = szAppName;

    RegisterClass (&wc);

    hwnd = CreateWindow (szAppName, szAppName,
        WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
        CW_USEDEFAULT, HWND_DESKTOP, NULL, hInstance, NULL);

    // Normally we'd call ShowWindow to display the window right here, but
    // leaving the window hidden allows RUPL to run invisibly.

    while (GetMessage (&msg, NULL, 0, 0)) {
        TranslateMessage (&msg);
        DispatchMessage (&msg);
    }
    return msg.wParam;
}

/////////////////////////////////////////////////////////////////////////////
// Window procedure

LRESULT CALLBACK WndProc (HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    static HHOOK hHook;
    HINSTANCE hInstDll;
    char szDir[MAX_PATH];
    ATOM atom;

    switch (msg) {

    case WM_CREATE:
        //
        // Create and initialize a global shell hook and initialize OLE.
        //
        hInstDll = (HINSTANCE) GetModuleHandle ("rupllib.dll");
        hHook = SetWindowsHookEx (WH_SHELL, (HOOKPROC) ShellProc,
            hInstDll, 0);

        if (hHook == NULL) {
            MessageBox (hwnd, "Unable to establish hook "\
                "(SetWindowsHookEx failed)", "Error", MB_ICONSTOP | MB_OK);
            return -1;
        }

        if (!SetHookHandle (hHook) || !SetWindowHandle (hwnd)) {
            MessageBox (hwnd, "Another instance of RUPL is running",
                "Error", MB_ICONSTOP | MB_OK);
            UnhookWindowsHookEx (hHook);
            return -1;
        }

        CoInitialize (NULL);
        return 0;

    case WM_USER:
        //
        // Add an item to the Recently Used menu. Use the directory name
        // passed through the global atom table as the working directory for
        // the application. Use the window handle passed in lParam to gather
        // other information about the application.
        //
        atom = (ATOM) wParam;
        GlobalGetAtomName (atom, szDir, sizeof (szDir));
        GlobalDeleteAtom (atom);
        UpdateMRUList (hwnd, (HWND) lParam, szDir);
        return 0;

    case WM_DESTROY:
        //
        // Remove the hook, uninitialize OLE, and terminate.
        //
        UnhookWindowsHookEx (hHook);
        CoUninitialize ();
        PostQuitMessage (0);
        return 0;
    }
    return DefWindowProc (hwnd, msg, wParam, lParam);
}

/////////////////////////////////////////////////////////////////////////////
// Helper functions

void UpdateMRUList (HWND hwnd, HWND hwndTarget, LPSTR pszDir)
{
    HWND hwndOwner;
    char szExeName[MAX_PATH];
    char szTitle[MAX_PATH];
    char szProgramsFolder[MAX_PATH];
    char szRecentlyUsedFolder[MAX_PATH];
    char szLnkName[MAX_PATH];
    UINT nMRUMax;

    // Make sure hwndTarget refers to an unowned window.
    while ((hwndOwner = GetWindow (hwndTarget, GW_OWNER)) != NULL)
        hwndTarget = hwndOwner;

    // Convert the window handle into a path name.
    if (!HwndToPath (hwndTarget, szExeName, sizeof (szExeName)))
        return;

    // Quit now if szExeName holds the EXE name of an app we don't support.
    if (!IsExeOK (szExeName))
        return;

    // Don't do anything if the window has a sibling that belongs to the
    // same process and a shortcut has already been created from the sibling.
    // This eliminates redundant menu entries for apps that create multiple
    // unowned top-level windows.
    if (IsDuplicate (hwndTarget, szExeName))
        return;

    // Generate a descriptive name for the application.
    if (!GetTitleFromRegistry (szExeName, szTitle, sizeof (szTitle))) {
        if (HwndToTitle (hwndTarget, szExeName, szTitle, sizeof (szTitle)))
            DeleteNonDescriptiveName (hwnd, szExeName);
        else {
            ExeToTitle (szExeName, szTitle, sizeof (szTitle));
            if (RefreshExistingLnkFile (hwnd, szExeName))
                return;
        }
    }

    // Get path names for the Programs folder and the Recently Used folder.
    if (!GetProgramsFolder (hwnd, szProgramsFolder) ||
        !GetRecentlyUsedFolder (hwnd, szRecentlyUsedFolder))
        return;

    // Create the Recently Used folder if it doesn't already exist and
    // build a path name for the shortcut that will be created in the folder.
    CreateDirectory (szRecentlyUsedFolder, NULL);

    lstrcpy (szLnkName, szRecentlyUsedFolder);
    lstrcat (szLnkName, "\\");
    lstrcat (szLnkName, szTitle);
    lstrcat (szLnkName, ".lnk");

    // Create a shortcut in the Recently Used folder. If the number of
    // shortcuts exceeds the maximum number of entries allowed in the menu,
    // delete one or more shortcuts.
    if (SUCCEEDED (CreateLink (szExeName, pszDir, szLnkName))) {
        nMRUMax = GetMRUMax ();
        while (GetLnkFileCount (szRecentlyUsedFolder) > nMRUMax)
            DeleteOldestLnkFile (szRecentlyUsedFolder);
        SHChangeNotify (SHCNE_UPDATEDIR, SHCNF_PATH, szProgramsFolder, 0);
    }
}

BOOL HwndToPath (HWND hwnd, LPSTR pszPath, int nSize)
{
    DWORD pid;
    HANDLE hSnapShot;
    PROCESSENTRY32 pe;
    BOOL bResult, r;
    int nLen;

    // Use the process ID to convert hwnd into an executable file name.
    GetWindowThreadProcessId (hwnd, &pid);
    hSnapShot = CreateToolhelp32Snapshot (TH32CS_SNAPPROCESS, pid);
    pe.dwSize = sizeof (pe);
    bResult = FALSE;

    for (r=Process32First (hSnapShot, &pe); r;
        r=Process32Next (hSnapShot, &pe)) {
        if (pe.th32ProcessID == pid) {
            if (pe.th32ParentProcessID != 0) {
                lstrcpy (pszPath, pe.szExeFile);
                bResult = TRUE;
            }
            break;
        }
    }

    CloseHandle (hSnapShot);

    // Fail the call if the executable file name doesn't end in ".EXE."
    if (bResult) {
        if (((nLen = lstrlen (pszPath)) < 5) ||
            (lstrcmpi (&pszPath[nLen - 4], ".exe") != 0))
            bResult = FALSE;
    }
    return bResult;
}

BOOL HwndToTitle (HWND hwnd, LPSTR pszExeName, LPSTR pszTitle, int nSize)
{
    int nLen;
    char szTemp[MAX_PATH];

    nLen = GetWindowText (hwnd, pszTitle, nSize);
    if ((nLen == 0) || !ValidateTitle (pszTitle))
        return FALSE;

    if (FindString (pszTitle, " - ") != -1) {
        if (nLen >= 11) {
            if (_strnicmp (pszTitle, "Untitled - ", 11) == 0) {
                lstrcpy (szTemp, pszTitle);
                lstrcpy (pszTitle, &szTemp[11]);
                return TRUE;
            }
            else if (lstrcmpi (&pszTitle[nLen - 11], " - Untitled") == 0) {
                pszTitle[nLen - 11] = 0;
                return TRUE;
            }
        }
        return FALSE;
    }
    return TRUE;
}

BOOL GetTitleFromRegistry (LPSTR pszExeName, LPSTR pszTitle, int nSize)
{
    HKEY hKey;
    BOOL bResult;
    char szRegKey[sizeof (g_szRegKey) + 32];
    DWORD dwSize;
    int i;

    lstrcpy (szRegKey, g_szRegKey);
    lstrcat (szRegKey, "\\Descriptive Names");
    for (i=lstrlen (pszExeName); pszExeName[i] != '\\'; i--);
    bResult = FALSE;

    if (RegOpenKeyEx (HKEY_LOCAL_MACHINE, szRegKey, 0,
        KEY_QUERY_VALUE, &hKey) == ERROR_SUCCESS) {
        dwSize = (DWORD) nSize;
        if (RegQueryValueEx (hKey, &pszExeName[i + 1], NULL, NULL,
            pszTitle, &dwSize) == ERROR_SUCCESS)
            bResult = TRUE;
        RegCloseKey (hKey);
    }
    return bResult;
}

BOOL ExeToTitle (LPSTR pszExeName, LPSTR pszTitle, int nSize)
{
    int i;

    for (i=lstrlen (pszExeName); pszExeName[i] != '\\'; i--);
    lstrcpy (pszTitle, &pszExeName[i + 1]);
    CharLower (pszTitle);
    CharUpperBuff (pszTitle, 1);
    return TRUE;
}

BOOL RefreshExistingLnkFile (HWND hwnd, LPSTR pszExeName)
{
    HANDLE hFind;
    WIN32_FIND_DATA fd;
    char szRecentlyUsedFolder[MAX_PATH];
    char szFileSpec[MAX_PATH];
    char szFileName[MAX_PATH];
    char szTemp[MAX_PATH];
    UINT nMRUMax;

    // First delete any extraneous LNK files from the Recently Used folder.
    if (!GetRecentlyUsedFolder (hwnd, szRecentlyUsedFolder))
        return FALSE;

    nMRUMax = GetMRUMax ();
    while (GetLnkFileCount (szRecentlyUsedFolder) > nMRUMax)
        DeleteOldestLnkFile (szRecentlyUsedFolder);

    // Now see if there's an existing LNK file targeting the same EXE. If
    // there is, update the file's date and time and return TRUE.
    lstrcpy (szFileSpec, szRecentlyUsedFolder);
    lstrcat (szFileSpec, "\\*.lnk");

    hFind = FindFirstFile (szFileSpec, &fd);
    if (hFind != INVALID_HANDLE_VALUE) {
        lstrcpy (szFileName, szRecentlyUsedFolder);
        lstrcat (szFileName, "\\");
        lstrcat (szFileName, fd.cFileName);
        ResolveLink (szFileName, szTemp, sizeof (szTemp));
        if (lstrcmpi (szTemp, pszExeName) == 0) {
            UpdateFileDateAndTime (szFileName);
            GetProgramsFolder (hwnd, szTemp);
            SHChangeNotify (SHCNE_UPDATEDIR, SHCNF_PATH, szTemp, 0);
            return TRUE;
        }
        while (FindNextFile (hFind, &fd)) {
            lstrcpy (szFileName, szRecentlyUsedFolder);
            lstrcat (szFileName, "\\");
            lstrcat (szFileName, fd.cFileName);
            ResolveLink (szFileName, szTemp, sizeof (szTemp));
            if (lstrcmpi (szTemp, pszExeName) == 0) {
                UpdateFileDateAndTime (szFileName);
                GetProgramsFolder (hwnd, szTemp);
                SHChangeNotify (SHCNE_UPDATEDIR, SHCNF_PATH, szTemp, 0);
                return TRUE;
            }
        }
    }
    return FALSE;
}

BOOL IsExeOK (LPSTR pszExeName)
{
    int i, nLen1, nLen2;
    char szRegKey[sizeof (g_szRegKey) + 32];
    BOOL bResult;
    HKEY hKey;

    static char *szBadNames[] = {
        "Winhlp32.exe",     // Windows help engine
        "Quikview.exe",     // Quick View engine
        "Infoview.exe",     // Infoviewer engine
        "Explorer.exe",     // Explorer
        "Spool32.exe",      // Print spooler
        "_Backgnd.exe",     // Lotus Notes background
        "Cid.exe",          // CompuServe Internet Dialer
        "Rundll32.exe",     // And the others...
        "Mprexe.exe",
        "Msgsrv32.exe",
        "Wsasrv.exe"
    };

    // First check the hardcoded list of EXEs.
    nLen1 = strlen (pszExeName);
    for (i=0; i<11; i++) {
        nLen2 = lstrlen (szBadNames[i]);
        if ((nLen1 >= nLen2) &&
            (lstrcmpi (szBadNames[i], &pszExeName[nLen1 - nLen2]) == 0))
            return FALSE;
    }

    // Then check the registry for excluded EXE names.
    lstrcpy (szRegKey, g_szRegKey);
    lstrcat (szRegKey, "\\Excluded Programs");
    for (i=nLen1; pszExeName[i] != '\\'; i--);
    bResult = TRUE;

    if (RegOpenKeyEx (HKEY_LOCAL_MACHINE, szRegKey, 0,
        KEY_QUERY_VALUE, &hKey) == ERROR_SUCCESS) {
        if (RegQueryValueEx (hKey, &pszExeName[i + 1], NULL, NULL,
            NULL, NULL) == ERROR_SUCCESS)
            bResult = FALSE;
        RegCloseKey (hKey);
    }
    return bResult;
}

int FindString (LPSTR pszString, LPSTR pszSubstring)
{
    int nResult, nStringLen, nSubstringLen, nCount, i;

    nStringLen = lstrlen (pszString);
    nSubstringLen = lstrlen (pszSubstring);
    nCount = nStringLen - nSubstringLen + 1;
    nResult = -1;

    if (nCount > 0) {
        for (i=0; i<nCount; i++) {
            if (_strnicmp (&pszString[i], pszSubstring, nSubstringLen) == 0) {
                nResult = i;
                break;
            }
        }
    }
    return nResult;
}

BOOL ValidateTitle (LPSTR pszText)
{
    int nLen, i, j;
    static char chBadChars[] = { '\\', '/', '"', '.', '*', '?' , '<', '>' };

    if ((nLen = lstrlen (pszText)) > 0) {
        for (i=0; i<nLen; i++) {
            for (j=0; j<sizeof (chBadChars) / sizeof (char); j++) {
                if (pszText[i] == chBadChars[j])
                    return FALSE;
            }
        }
    }
    return TRUE;
}

BOOL IsDuplicate (HWND hwnd, LPSTR pszExeName)
{
    HWND h;
    DWORD pid1, pid2;
    char szTitle[MAX_PATH];
    char szLnkName[MAX_PATH];
    WIN32_FIND_DATA fd;
    HANDLE hFind;

    GetWindowThreadProcessId (hwnd, &pid1);
    for (h=GetWindow (hwnd, GW_HWNDFIRST); h; h=GetWindow (h, GW_HWNDNEXT)) {
        if ((h != hwnd) && (GetWindow (h, GW_OWNER) == NULL) &&
            IsWindowVisible (h)) {
            GetWindowThreadProcessId (h, &pid2);
            if (pid1 == pid2) {
                HwndToTitle (h, pszExeName, szTitle, sizeof (szTitle));
                GetRecentlyUsedFolder (hwnd, szLnkName);
                lstrcat (szLnkName, "\\");
                lstrcat (szLnkName, szTitle);
                lstrcat (szLnkName, ".lnk");
                hFind = FindFirstFile (szLnkName, &fd);
                if (hFind != INVALID_HANDLE_VALUE) {
                    CloseHandle (hFind);
                    return TRUE;
                }
            }
        }
    }
    return FALSE;
}

BOOL GetProgramsFolder (HWND hwnd, LPSTR pszPath)
{
    BOOL bResult;
    LPITEMIDLIST pidl;
    LPMALLOC pMalloc;

    bResult = FALSE;
    if (SUCCEEDED (SHGetMalloc (&pMalloc))) {
        if (SUCCEEDED (SHGetSpecialFolderLocation (hwnd, CSIDL_PROGRAMS,
            &pidl))) {
            if (SHGetPathFromIDList (pidl, pszPath))
                bResult = TRUE;
            pMalloc->lpVtbl->Free (pMalloc, pidl);
        }
        pMalloc->lpVtbl->Release (pMalloc);
    }
    return bResult;
}

BOOL GetRecentlyUsedFolder (HWND hwnd, LPSTR pszPath)
{
    if (!GetProgramsFolder (hwnd, pszPath))
        return FALSE;

    if (pszPath[lstrlen (pszPath) - 1] != '\\')
        lstrcat (pszPath, "\\");
    lstrcat (pszPath, "Recently Used");
    return TRUE;
}

HRESULT CreateLink (LPSTR pszFileName, LPSTR pszWorkingDir, LPSTR pszLnkName)
{
    HRESULT hResult;
    WORD szWide[MAX_PATH];
    IPersistFile *ppf;
    IShellLink *psl;

    hResult = CoCreateInstance (&CLSID_ShellLink, NULL,
        CLSCTX_INPROC_SERVER, &IID_IShellLink, &psl);

    if (SUCCEEDED (hResult)) {
        psl->lpVtbl->SetPath (psl, pszFileName);
        psl->lpVtbl->SetWorkingDirectory (psl, pszWorkingDir);

        hResult = psl->lpVtbl->QueryInterface (psl, &IID_IPersistFile, &ppf);
        if (SUCCEEDED (hResult)) {
            MultiByteToWideChar (CP_ACP, 0, pszLnkName, -1, szWide,
                MAX_PATH);
            hResult = ppf->lpVtbl->Save (ppf, szWide, TRUE);
            ppf->lpVtbl->Release (ppf);
        }
        psl->lpVtbl->Release (psl);
    }
    return hResult;
}

HRESULT ResolveLink (LPSTR pszLnkName, LPSTR pszTarget, int nSize)
{
    HRESULT hResult;
    IPersistFile *ppf;
    IShellLink *psl;
    WORD szWide[MAX_PATH];
    WIN32_FIND_DATA fd;

    hResult = CoCreateInstance (&CLSID_ShellLink, NULL,
        CLSCTX_INPROC_SERVER, &IID_IShellLink, &psl);

    if (SUCCEEDED (hResult)) {
        hResult = psl->lpVtbl->QueryInterface (psl, &IID_IPersistFile, &ppf);
        if (SUCCEEDED (hResult)) {
            MultiByteToWideChar (CP_ACP, 0, pszLnkName, -1, szWide,
                MAX_PATH);
            hResult = ppf->lpVtbl->Load (ppf, szWide, STGM_READ);
            if (SUCCEEDED (hResult)) {
                hResult = psl->lpVtbl->GetPath (psl, pszTarget, nSize, &fd,
                    SLGP_UNCPRIORITY);
            }
            ppf->lpVtbl->Release (ppf);
        }
        psl->lpVtbl->Release (psl);
    }
    return hResult;
}

UINT GetLnkFileCount (LPSTR pszPath)
{
    UINT nCount;
    char szFileSpec[MAX_PATH];
    WIN32_FIND_DATA fd;
    HANDLE hFind;

    nCount = 0;
    lstrcpy (szFileSpec, pszPath);
    lstrcat (szFileSpec, "\\*.lnk");

    hFind = FindFirstFile (szFileSpec, &fd);
    if (hFind != INVALID_HANDLE_VALUE) {
        nCount++;
        while (FindNextFile (hFind, &fd))
            nCount++;
        FindClose (hFind);
    }
    return nCount;
}

void DeleteOldestLnkFile (LPSTR pszPath)
{
    WIN32_FIND_DATA fd;
    HANDLE hFile, hFind;
    char szFileSpec[MAX_PATH];
    char szCurrentFile[MAX_PATH];
    char szOldestFile[MAX_PATH];
    FILETIME ft, ftOldest;
    int nIndex;

    lstrcpy (szFileSpec, pszPath);
    lstrcat (szFileSpec, "\\*.lnk");
    lstrcpy (szCurrentFile, pszPath);
    lstrcat (szCurrentFile, "\\");
    nIndex = lstrlen (szCurrentFile);

    hFind = FindFirstFile (szFileSpec, &fd);
    if (hFind != INVALID_HANDLE_VALUE) {
        lstrcpy (&szCurrentFile[nIndex], fd.cFileName);
        lstrcpy (szOldestFile, szCurrentFile);
        hFile = CreateFile (szCurrentFile, GENERIC_READ, 0, NULL,
            OPEN_EXISTING, 0, NULL);
        GetFileTime (hFile, NULL, NULL, &ftOldest);
        CloseHandle (hFile);

        while (FindNextFile (hFind, &fd)) {
            lstrcpy (&szCurrentFile[nIndex], fd.cFileName);
            hFile = CreateFile (szCurrentFile, GENERIC_READ, 0, NULL,
                OPEN_EXISTING, 0, NULL);
            GetFileTime (hFile, NULL, NULL, &ft);
            if (CompareFileTime (&ft, &ftOldest) < 0) {
                ftOldest.dwLowDateTime = ft.dwLowDateTime;
                ftOldest.dwHighDateTime = ft.dwHighDateTime;
                lstrcpy (szOldestFile, szCurrentFile);
            }
            CloseHandle (hFile);
        }
        DeleteFile (szOldestFile);
    }
}

void DeleteNonDescriptiveName (HWND hwnd, LPSTR pszExeName)
{
    int i;
    char szLnkName[MAX_PATH];

    for (i=lstrlen (pszExeName); pszExeName[i] != '\\'; i--);
    GetRecentlyUsedFolder (hwnd, szLnkName);
    lstrcat (szLnkName, "\\");
    lstrcat (szLnkName, &pszExeName[i + 1]);
    lstrcat (szLnkName, ".lnk");

    DeleteFile (szLnkName);
}

void UpdateFileDateAndTime (LPSTR pszPath)
{
    BYTE byte;
    DWORD dwCount;
    HANDLE hFile;

    hFile = CreateFile (pszPath, GENERIC_READ | GENERIC_WRITE, 0, NULL,
        OPEN_EXISTING, 0, NULL);

    if (hFile != INVALID_HANDLE_VALUE) {
        ReadFile (hFile, &byte, 1, &dwCount, NULL);
        SetFilePointer (hFile, 0, NULL, FILE_BEGIN);
        WriteFile (hFile, &byte, 1, &dwCount, NULL);
        CloseHandle (hFile);
    }
}

UINT GetMRUMax ()
{
    HKEY hKey;
    DWORD dwSize;
    UINT nMRUMax;

    nMRUMax = 15;
    if (RegOpenKeyEx (HKEY_LOCAL_MACHINE, g_szRegKey, 0,
        KEY_QUERY_VALUE, &hKey) == ERROR_SUCCESS) {

        dwSize = sizeof (int);
        RegQueryValueEx (hKey, "MaxMRU", NULL, NULL, (PBYTE) &nMRUMax,
            &dwSize);
        RegCloseKey (hKey);

        // Sanity check
        nMRUMax = max (2, nMRUMax);
        nMRUMax = min (128, nMRUMax);
    }
    return nMRUMax;
}
